#include <iostream>
#include <stack>
#include <map>
#include <cstring>

#include "Profile.h"

using namespace std;

//paired:0
//single strand: 1
double s2r(char c){
    if(c=='<' || c=='>' || islower(c) || isupper(c))
        return 0.;
    else
        return 1.;
}

Profile::Profile(const char* structure, double** matrix, int length){
    int k=0;
    map<char, stack<int> > pair_stack;		//multiple stacks for all possible pairs (<>, Aa)
	
	//set structure in Profile
    this->structure=new char[length+1];
    strcpy(this->structure, structure);

	this->length=length;	//set length in Profile
	 
    p_react=new int[length];
    PSSM=new double*[length];
    basepair=new int[length];
    memset(basepair, -1, sizeof(int)*length);
    
    for(int i=0; i<length; i++){
        PSSM[i]=new double[6];
        p_react[i]=s2r(structure[i]);	//set p_react in Profile
        
        for(int j=0; j<6; j++){			//set PSSM in Profile
            PSSM[i][j]=matrix[i][j];
        }

        if(structure[i]=='<'){
            if(pair_stack.find('<')==pair_stack.end())
                pair_stack.insert( pair<char, stack<int> >('<', stack<int>()) );
            (pair_stack.find('<')->second).push(i);
        }
        else if(structure[i]=='>'){
            k=(pair_stack.find('<')->second).top();
            (pair_stack.find('<')->second).pop();
            basepair[i]=k;
            basepair[k]=i;
        }
        else if(isupper(structure[i])){
            if(pair_stack.find(structure[i])==pair_stack.end())
                pair_stack.insert( pair<char, stack<int> >(structure[i], stack<int>()) );
            (pair_stack.find(structure[i])->second).push(i);
        }
        else if(islower(structure[i])){
            k=(pair_stack.find(toupper(structure[i]))->second).top();
            (pair_stack.find(toupper(structure[i]))->second).pop();
            basepair[i]=k;
            basepair[k]=i;
        }

    }
}

Profile::~Profile(){
    delete[] structure;
    delete[] p_react;
    delete[] basepair;
    for(int i=0; i<length; i++)
        delete[] PSSM[i];
}
