#include <iostream>
#include <stack>
#include <vector>
#include <algorithm>
#include <string>
#include <cmath>

#include <cstdio>
#include <cstring>

#include "algorithm.h"
#include "param.h"
#include "stats.h"

using namespace std;

double BASEPAIR[6][6]={
    {0.,  0.,  0.,  1.,  0., 0.},
    {0.,  0.,  1.,  0.,  0., 0.},
    {0.,  1.,  0.,  .5,  0., 0.},
    {1.,  0.,  .5,  0.,  0., 0.},
    {0.,  0.,  0.,  0.,  0., 0.},
    {0.,  0.,  0.,  0.,  0., 0.}
};

//universal match score for all protocols
double match_score_uni(Profile* P, Target* T, int i, int j){
    if(T->react[j]==-999.)
        return 0.;
    else if(P->p_react[i]==1. && T->react[j]>=react_c)
        return 1.;
    else if(P->p_react[i]==0. && T->react[j]<react_c)
        return 1.;
    else
        return -1.;
}

//match score for shape protocols
double match_score_shape(Profile* P, Target* T, int i, int j){
	double p1, p2;
	if(T->react[j]==-999.){
		return 0.;
	}
	else{
		p1=dexpon(T->react[j]);		//probability of unpairing at Target
		p2=dgev(T->react[j]);		//probability of pairing at Target 
		if(P->p_react[i]==1.)		//unpaired at profile
			return log10(p1/p2);
		else if(P->p_react[i]==0.)	//paired at profile
			return log10(p2/p1);
	}
	return 0.;
}

char TBM_entry(double M, double M1, double M2, double M3, double IX, double IX1, double IX2, double IY, double IY1, double IY2){
    char ret=(char)0;

    if(M==M1)
        ret|=(char)64;
    if(M==M2)
        ret|=(char)32;
    if(M==M3)
        ret|=(char)16;
    if(IX==IX1)
        ret|=(char)8;
    if(IX==IX2)
        ret|=(char)4;
    if(IY==IY1)
        ret|=(char)2;
    if(IY==IY2)
        ret|=(char)1;
    return ret;
}

bool test_TBM_entry(char entry, int i){
    switch(i){
    case 6:
        return (entry & char(64))==(char)64;
    case 5:
        return (entry & char(32))==(char)32;
    case 4:
        return (entry & char(16))==(char)16;
    case 3:
        return (entry & char(8))==(char)8;
    case 2:
        return (entry & char(4))==(char)4;
    case 1:
        return (entry & char(2))==(char)2;
    case 0:
        return (entry & char(1))==(char)1;
    default:
        return false;
    }
}

double* SG_affine_align(Profile* P, Target* T, char** TBM){
    int PL;
    int TL;
    double **M, **IX, **IY;
    double* sco;

    PL=P->length;
    TL=T->length;

    M=new double*[PL+1];
    IX=new double*[PL+1];
    IY=new double*[PL+1];
    for(int i=0; i<=PL; i++){
        M[i]=new double[TL+1];
        IX[i]=new double[TL+1];
        IY[i]=new double[TL+1];
        for(int j=0; j<=TL; j++){
            M[i][j]=NEGATIVE_INFINITY;
            IX[i][j]=NEGATIVE_INFINITY;
            IY[i][j]=NEGATIVE_INFINITY;
        }
    }

    sco=new double[TL+1];

    TBM[0][0]=(char)128;

    for(int i=0; i<=PL; i++){
        M[i][0]=gap_o+gap_e*i;
        TBM[i][0]=(char)128;
    }
    for(int j=0; j<=TL; j++){
        M[0][j]=0.;
        TBM[0][j]=(char)128;
    }
    M[0][0]=0.;

    double M1, M2, M3, IX1, IX2, IY1, IY2;
    for(int i=1; i<=PL; i++){
        for(int j=1; j<=TL; j++){
            IX1=M[i-1][j] + gap_o + gap_e + beta*P->PSSM[i-1][5];
            IX2=IX[i-1][j] + gap_e + beta*P->PSSM[i-1][5];
            IX[i][j]=max(IX1, IX2);

            IY1=M[i][j-1] + gap_o + gap_e;
            IY2=IY[i][j-1] + gap_e;
            IY[i][j]=max(IY1, IY2);

            if(sflag==0)
				M1=M[i-1][j-1] + alpha*match_score_uni(P, T, i-1, j-1) + beta*P->PSSM[i-1][T->nseq[j-1]];
            else
				M1=M[i-1][j-1] + alpha*match_score_shape(P, T, i-1, j-1) + beta*P->PSSM[i-1][T->nseq[j-1]];
            M2=IX[i][j];
            M3=IY[i][j];
            M[i][j]=max(M1, max(M2, M3));

            TBM[i][j]=TBM_entry(M[i][j], M1, M2, M3, IX[i][j], IX1, IX2, IY[i][j], IY1, IY2);

        }
    }

    memcpy(sco, M[PL], (TL+1)*sizeof(double));
    for(int i=0; i<=PL; i++){
        delete[] M[i];
        delete[] IX[i];
        delete[] IY[i];
    }
    return sco;
}

//import: Te is the position in TBM
void SG_affine_traceback(char** TBM, int PL, int Te, vector<vector<int> > &ALNs, vector<int> &ALN_starts){
//type: M: 2, IX: 1, IY: 0   what type of function used at pos
    stack<int> type_stack;
//next_pos: i, j
    stack<int> pos_stack;
//alignment stack
    stack<vector<int> > ALN_stack;
/*
X   2   X   1   -   0
Y       -       Y
*/
    vector<int> cur_ALN;

    int max_ret_aln=20;

    type_stack.push(2);
    pos_stack.push(PL);
    pos_stack.push(Te);
    ALN_stack.push(cur_ALN);

    int i,j,type;
    while(ALN_stack.empty()!=true){
        cur_ALN=ALN_stack.top();
        ALN_stack.pop();

        j=pos_stack.top();
        pos_stack.pop();

        i=pos_stack.top();
        pos_stack.pop();

        type=type_stack.top();
        type_stack.pop();

        if(i==0 || j==0){
            if(j==0)
                for(int k=0; k<i; k++) cur_ALN.push_back(1);
            reverse(cur_ALN.begin(), cur_ALN.end());
            ALNs.push_back(cur_ALN);
            ALN_starts.push_back(j);

            max_ret_aln--;
            if(max_ret_aln==0) break;

            continue;
        }
        if(type==2){        //type M
            if(test_TBM_entry(TBM[i][j], 5)){       //M2
                //cur_ALN.push_back(1);
                type_stack.push(1);
                pos_stack.push(i);
                pos_stack.push(j);
                ALN_stack.push(cur_ALN);
            }
            if(test_TBM_entry(TBM[i][j], 4)){       //M3
                //cur_ALN.push_back(0);
                type_stack.push(0);
                pos_stack.push(i);
                pos_stack.push(j);
                ALN_stack.push(cur_ALN);
            }
            if(test_TBM_entry(TBM[i][j], 6)){       //M1
                cur_ALN.push_back(2);
                type_stack.push(2);
                pos_stack.push(i-1);
                pos_stack.push(j-1);
                ALN_stack.push(cur_ALN);
            }
        }
        if(type==1){    //type IX
            cur_ALN.push_back(1);
            if(test_TBM_entry(TBM[i][j], 3)){       //IX1
                type_stack.push(2);
                pos_stack.push(i-1);
                pos_stack.push(j);
                ALN_stack.push(cur_ALN);
            }
            if(test_TBM_entry(TBM[i][j], 2)){       //IX2
                type_stack.push(1);
                pos_stack.push(i-1);
                pos_stack.push(j);
                ALN_stack.push(cur_ALN);
            }
        }
        if(type==0){    //type IY
            cur_ALN.push_back(0);
            if(test_TBM_entry(TBM[i][j], 1)){   //IY1
                type_stack.push(2);
                pos_stack.push(i);
                pos_stack.push(j-1);
                ALN_stack.push(cur_ALN);
            }
            if(test_TBM_entry(TBM[i][j], 0)){   //IY2
                type_stack.push(0);
                pos_stack.push(i);
                pos_stack.push(j-1);
                ALN_stack.push(cur_ALN);
            }
        }

    }
    return;
}

void show_ALN(Profile* P, const char* seq, double* reacts, int s, vector<int> ALN){
    int pi,ti;
    pi=0;
    ti=s;

    string s_aln;   //structure line e.g. <<<...>>>
    string p_aln;   //profile line e.g. aCgtACGt
    string a_aln;   //alignment line e.g. (((..||..)))
    string t_aln;   //target line e.g. ACGTACGT
    string r_aln;   //react line e.g. 01010111

    char char_in_p; //char in profile decided by the ratio of each nucleotide

    int len=ALN.size();

    //mapping "<" in profile to the its position in the alignment
    int mapping[P->length];
    memset(mapping, -1, P->length*sizeof(int));

    //for finding the nucleotide with maximum ratio in each position of profile
    int max_j;
    double max_v;

    for(vector<int>::iterator it=ALN.begin(); it!=ALN.end(); ++it){
        if(*it==2){
            s_aln+=P->structure[pi];

            max_j=0;
            max_v=NEGATIVE_INFINITY;
            for(int j=0; j<6; j++){
                if(P->PSSM[pi][j]>max_v){
                    max_v=P->PSSM[pi][j];
                    max_j=j;
                }
            }

            if(max_j==0 && P->PSSM[pi][max_j]>=0.5) char_in_p='A';
            else if(max_j==0 && P->PSSM[pi][max_j]<0.5) char_in_p='a';
            else if(max_j==1 && P->PSSM[pi][max_j]>=0.5) char_in_p='C';
            else if(max_j==1 && P->PSSM[pi][max_j]<0.5) char_in_p='c';
            else if(max_j==2 && P->PSSM[pi][max_j]>=0.5) char_in_p='G';
            else if(max_j==2 && P->PSSM[pi][max_j]<0.5) char_in_p='g';
            else if(max_j==3 && P->PSSM[pi][max_j]>=0.5) char_in_p='U';
            else if(max_j==3 && P->PSSM[pi][max_j]<0.5) char_in_p='u';
            else if(max_j==4) char_in_p='-';
            else if(max_j==5 && P->PSSM[pi][max_j]>=0.5) char_in_p='X';
            else if(max_j==5 && P->PSSM[pi][max_j]<0.5) char_in_p='x';

            p_aln+=char_in_p;

            t_aln+=seq[ti];

			if(sflag==0){
				if(reacts[ti]>=react_c)
					r_aln+="1";
				else if(reacts[ti]==-999.)
					r_aln+="=";
				else
					r_aln+="0";
			}
			else{
				if(dexpon(reacts[ti])>=dgev(reacts[ti]))
					r_aln+="1";
				else if(reacts[ti]==-999.)
					r_aln+="=";
				else
					r_aln+="0";
			}
                
            if(P->basepair[pi]==-1 && toupper(char_in_p)==toupper(seq[ti])) a_aln+="|";
            else if(P->basepair[pi]==-1 && toupper(char_in_p)!=toupper(seq[ti])) a_aln+=" ";
            else if(P->basepair[pi]!=-1 && pi<P->basepair[pi]) {a_aln+=" "; mapping[pi]=distance(ALN.begin(), it);}
            else if(P->basepair[pi]!=-1 && pi>P->basepair[pi] && mapping[P->basepair[pi]]==-1) a_aln+=" ";
            else if(P->basepair[pi]!=-1 && pi>P->basepair[pi] && mapping[P->basepair[pi]]!=-1) {
                if(BASEPAIR[c2n(seq[ti])][c2n(t_aln[mapping[P->basepair[pi]]])]!=0){
                    a_aln+=")";
                    a_aln.replace(mapping[P->basepair[pi]], 1, "(");
                }
                else
                    a_aln+=" ";
            }
            pi++;
            ti++;
        }
        else if(*it==1){
            s_aln+=P->structure[pi];
            max_j=0;
            max_v=NEGATIVE_INFINITY;
            for(int j=0; j<6; j++){
                if(P->PSSM[pi][j]>max_v){
                    max_v=P->PSSM[pi][j];
                    max_j=j;
                }
            }

            if(max_j==0 && P->PSSM[pi][max_j]>=0.5) char_in_p='A';
            else if(max_j==0 && P->PSSM[pi][max_j]<0.5) char_in_p='a';
            else if(max_j==1 && P->PSSM[pi][max_j]>=0.5) char_in_p='C';
            else if(max_j==1 && P->PSSM[pi][max_j]<0.5) char_in_p='c';
            else if(max_j==2 && P->PSSM[pi][max_j]>=0.5) char_in_p='G';
            else if(max_j==2 && P->PSSM[pi][max_j]<0.5) char_in_p='g';
            else if(max_j==3 && P->PSSM[pi][max_j]>=0.5) char_in_p='U';
            else if(max_j==3 && P->PSSM[pi][max_j]<0.5) char_in_p='u';
            else if(max_j==4) char_in_p='-';
            else if(max_j==5 && P->PSSM[pi][max_j]>=0.5) char_in_p='X';
            else if(max_j==5 && P->PSSM[pi][max_j]<0.5) char_in_p='x';

            p_aln+=char_in_p;
            a_aln+=" ";
            t_aln+='-';
            r_aln+='-';

            pi++;
        }
        else if(*it==0){
            s_aln+='-';
            p_aln+='-';
            a_aln+=' ';
            t_aln+=seq[ti];
            if(reacts[ti]>=react_c)
                r_aln+="1";
			else if(reacts[ti]==-999.)
				r_aln+="=";
            else
                r_aln+="0";
            ti++;
        }
    }

    int i=0;
    while(i<len){
        cout<<s_aln.substr(i, min(NUCL_PER_LINE, len-i))<<endl;
        cout<<p_aln.substr(i, min(NUCL_PER_LINE, len-i))<<endl;
        cout<<a_aln.substr(i, min(NUCL_PER_LINE, len-i))<<endl;
        cout<<t_aln.substr(i, min(NUCL_PER_LINE, len-i))<<endl;
        cout<<r_aln.substr(i, min(NUCL_PER_LINE, len-i))<<endl;
        cout<<endl;

        i+=NUCL_PER_LINE;
    }

    return;
}


double pair_bonus(Profile* P, Target* T, vector<vector<int> > ALNs, vector<int> ALN_starts, int* opt_ALN_i){
    double bonus, max_bonus;
    int bi, bj;
    int i,j;
    int* mapping=new int[P->length];

    max_bonus=0.;
    *opt_ALN_i=0;

    vector<int>::iterator s_it=ALN_starts.begin();
    for(vector<vector<int> >::iterator aln_it=ALNs.begin(); aln_it!=ALNs.end(); ++aln_it, ++s_it){
        bonus=0.;

        i=0;
        j=*s_it;
        memcpy(mapping, P->basepair, P->length*sizeof(int));
        for(vector<int>::iterator it=aln_it->begin(); it!=aln_it->end(); ++it){
            if(*it==2){
                if(mapping[i]>=0){
                    mapping[i]=j;
                }
                i++;
                j++;
            }
            else if(*it==1){
                mapping[i]=-1;
                i++;
            }
            else if(*it==0){
                j++;
            }
        }
//get mapping_
        for(int i=0; i<P->length; i++){
            if(P->basepair[i]>=0){
                bi=i;
                bj=P->basepair[i];
                if(bi<bj && mapping[bi]>=0 && mapping[bj]>=0) bonus+=BASEPAIR[T->nseq[mapping[bi]]][T->nseq[mapping[bj]]];
            }
        }
        if(bonus>max_bonus){
            max_bonus=bonus;
            *opt_ALN_i=distance(ALNs.begin(), aln_it);
        }
    }
    delete[] mapping;
    return max_bonus;
}


int max_sco(double* sco, int length){
    double max_sco=NEGATIVE_INFINITY;
    double ret=-1;
    for(int i=0; i<length; i++){
        if(sco[i]>max_sco){
            ret=i;
            max_sco=sco[i];
        }
    }
    return ret;
}
