#include <cmath>
#include <iostream>

#include "stats.h"

using namespace std;

double lngamma(double z){
    double x = 0;
    x += 0.1659470187408462e-06 / (z+7);
    x += 0.9934937113930748e-05 / (z+6);
    x -= 0.1385710331296526 / (z+5);
    x += 12.50734324009056 / (z+4);
    x -= 176.6150291498386 / (z+3);
    x += 771.3234287757674 / (z+2);
    x -= 1259.139216722289 / (z+1);
    x += 676.5203681218835 / z;
    x += 0.9999999999995183;
    return log(x) - 5.58106146679532777 - z + (z-0.5) * log(z+6.5);
}

#define GAMMA_EPS 1e-14
#define TINY 1e-290
double lgammainc(double s, double z)
{
        double sum, x;
        int k;
        for (k = 1, sum = x = 1.; k < 100; ++k) {
                sum += (x *= z / (s + k));
                if (x / sum < GAMMA_EPS) break;
        }
        return exp(s * log(z) - z - lngamma(s + 1.) + log(sum));
}

double ugammainc(double s, double z)
{
        int j;
        double C, D, f;
        f = 1. + z - s; C = f; D = 0.;
        for (j = 1; j < 100; ++j) {
                double a = j * (s - j), b = (j<<1) + 1 + z - s, d;
                D = b + a * D;
                if (D < TINY) D = TINY;
                C = b + a / C;
                if (C < TINY) C = TINY;
                D = 1. / D;
                d = C * D;
                f *= d;
                if (fabs(d - 1.) < GAMMA_EPS) break;
        }
        return exp(s * log(z) - z - lngamma(s) - log(f));
}

double dgamma(double s, double z){
    return z <= 1. || z < s ? lgammainc(s, z) : 1-ugammainc(s, z);
}

double pvalue_gamma(double x, double shape, double loc, double scale){
    return isnan(dgamma(shape, (x-loc)/scale)) ? 1.:1-dgamma(shape, (x-loc)/scale);
    //return 1.-dgamma(shape, (x-loc)/scale);
}

double dgev(double x, double shape, double loc, double scale){
	double temp=1+shape*(x-loc)/scale;
	return temp>0? 1./scale*pow(temp, -1.-1./shape)*exp(-1.*pow(temp, -1./shape)) : 0;
}

double dexpon(double x, double lambda){
	return x>=0.? lambda*exp(-1.*lambda*x) : 0;
}
