#!/usr/bin/python
from sys import argv, exit
from Bio import SeqIO
from getopt import getopt, GetoptError
from os import system, remove, path

import scipy.stats as ss
import scipy as sp

#A, C, G, U, X, -
def usage():
	print "./Rfam_Profile.py <-o gap_open> <-e gap_extend> <-a alpha> <-b beta> <-r react cutoff> <-p pvalue cutoff> <-s> rfam.stockholm seq seq_react"

CUTOFF=.5

try:
	opts, args=getopt(argv[1:], "o:e:a:b:r:p:s")
except GetoptError as err:
	print str(err)
	usage()	
	exit()

if len(args)!=3:
	usage()
	exit()


#get the consensus structure in a stockholm file
def Rfam_get_con_struct(fn):
	struct=''
	fp=open(fn)
	for line in fp.readlines():
		if line[:12]=='#=GC SS_cons':
			struct+=line[12:].strip()
	fp.close()
	return struct

#get the pairs in an RNA structure(Rfam format)
#considering pesudo knot
def Rfam_get_pairs(struct, pk=True):
	stack=[]
	if pk==True:
		stack_pk={}
	
	pairs=[-1]*len(struct)
	for i, c in enumerate(struct):
		if c=='<':
			stack.append(i)
		if c=='>':
			j=stack.pop()
			pairs[i]=j
			pairs[j]=i
		if pk==True:
			if c.isupper():
				if c not in stack_pk:
					stack_pk[c]=[]
				stack_pk[c].append(i)
			if c.islower():
				j=stack_pk[c.upper()].pop()
				pairs[i]=j
				pairs[j]=i
	return pairs


ALN={}
fn=args[0]
for record in SeqIO.parse(fn, "stockholm"):
	L=len(record.seq)
	ALN[record.id]=str(record.seq)

struct=Rfam_get_con_struct(fn)

pairs=Rfam_get_pairs(struct)

bases=[]
for i, v in enumerate(pairs):
	if v==-1:
		bases.append(i)

profile=['']*len(struct)
removed=[]
for i in bases:
	nA=0
	nC=0
	nG=0
	nU=0
	nX=0
	nI=0
	for s in ALN:
		if ALN[s][i]=='A':
			nA+=1
		elif ALN[s][i]=='C':
			nC+=1
		elif ALN[s][i]=='G':
			nG+=1
		elif ALN[s][i]=='U':
			nU+=1
		elif ALN[s][i]=='-':
			nI+=1
		else:
			nX+=1
	S=nA+nC+nG+nU+nI+nX
	if 1.*nI/S>CUTOFF:
		removed.append(i)
	profile[i]=(1.*nA/S, 1.*nC/S, 1.*nG/S, 1.*nU/S, 1.*nX/S, 1.*nI/S)

for i, j in enumerate(pairs):
	if profile[i]!='' or profile[j]!='':
		continue	
	nAi=0
	nCi=0
	nGi=0
	nUi=0
	nIi=0
	nXi=0
	
	nAj=0
	nCj=0
	nGj=0
	nUj=0
	nIj=0
	nXj=0
	for s in ALN:
		if ALN[s][i]=='A':
			nAi+=1
		elif ALN[s][i]=='C':
			nCi+=1
		elif ALN[s][i]=='G':
			nGi+=1
		elif ALN[s][i]=='U':
			nUi+=1
		elif ALN[s][i]=='-':
			nIi+=1
		else:
			nXi+=1
			
		if ALN[s][j]=='A':
			nAj+=1
		elif ALN[s][j]=='C':
			nCj+=1
		elif ALN[s][j]=='G':
			nGj+=1
		elif ALN[s][j]=='U':
			nUj+=1
		elif ALN[s][j]=='-':
			nIj+=1
		else:
			nXj+=1
	Si=nAi+nCi+nGi+nUi+nIi+nXi
	profile[i]=(1.*nAi/Si, 1.*nCi/Si, 1.*nGi/Si, 1.*nUi/Si, 1.*nXi/Si, 1.*nIi/Si)
	Sj=nAj+nCj+nGj+nUj+nIj+nXj
	profile[j]=(1.*nAj/Sj, 1.*nCj/Sj, 1.*nGj/Sj, 1.*nUj/Sj, 1.*nXj/Sj, 1.*nIj/Sj)
	if 1.*nIi/Si>CUTOFF or 1.*nIj/Sj>CUTOFF:
		removed.append(i)
		removed.append(j)


struct=''.join(c for i, c in enumerate(struct) if i not in removed)
profile=[p for i, p in enumerate(profile) if i not in removed]

ret=[]
ret.append(struct)
for p in profile:
	ret.append('\t'.join(map(lambda x: str(x), p)))

fp=open(path.basename(args[0])+'.profile.1', 'w')
for i in ret:
	fp.write(i+'\n')
fp.close()

system('../ProbeAlign -t %s %s %s %s > %s' % (' '.join(map(lambda x: ' '.join(x), opts)), path.basename(args[0])+'.profile.1', args[1], args[2], path.basename(args[0])+'.profile.2'))

data=[]
fp=open(path.basename(args[0])+'.profile.2')
for line in fp.readlines():
	data.append(float(line.strip()))
fp.close()

data=filter(lambda x: x>0, data)

param_gamma=ss.gamma.fit(data)

print '%f\t%f\t%f' % (param_gamma)
for i in ret:
	print i

if path.isfile(path.basename(args[0]+'.profile.1')):
	remove(path.basename(args[0]+'.profile.1'))
if path.isfile(path.basename(args[0]+'.profile.2')):
	remove(path.basename(args[0]+'.profile.2'))
